package PleskStructure;

use PleskVersion;
use XmlNode;

# IP address is the pair (address, type)

my %ipType; #ip -> 'shared'|'exclusive'
my %clientIpPool; #clientName -> @ip
my %clientDomains; #clientName -> domainName
my %domainIp; #domainName -> ip
my $defaultIp;

my $dbh;

## public interface

sub init {
  $dbh = shift;

  if (PleskVersion::atLeast(6, 0, 0)) {
    __init60();
  } elsif (PleskVersion::atLeast(5, 0, 0)) {
    __init50();
  } else {
    __init25();
  }
}

sub getClients {
  return keys %clientIpPool;
}

sub getDomains {
  return keys %domainIp;
}

sub getDomainsForClient {
  my ($clientName) = @_;
  return @{$clientDomains{$clientName}};
}

sub getClientIps {
  my ($clientName) = @_;
  return @{$clientIpPool{$clientName}};
}

sub getDomainIp {
  my ($domainName) = @_;
  return $domainIp{$domainName};
}

sub isExclusiveIp {
  my ($ip) = @_;

  return $ipType{$ip} eq "exclusive";
}

sub makeIpNode {
  my ($ip) = @_;
  return XmlNode->new('ip', 'children' => [XmlNode->new('ip-type', content => $ipType{$ip}),
                                           XmlNode->new('ip-address', content => $ip)]);
}

## Plesk 2.5

sub __init25 {
  my $defaultIp = __getDefaultIp();
  __populateIpsFromDns();

  my $sql = 'SELECT d.name, r.val FROM dns_recs r, domains d WHERE r.dom_id = d.id AND r.type = "A"';
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($domainName, $ipAddress) = @{$row};
    $ipAddress ||= $defaultIp;
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  __populateClientsFromDomains();
}

## Plesk 5.0, 5.5

sub __init50 {
  my $defaultIp = __getDefaultIp();
  __populateIpsFromDns();
  __populateIpsFromList("SELECT DISTINCT(ip_address) FROM hosting");

  my $sql = "SELECT d.name, h.ip_address FROM domains d LEFT JOIN hosting h ON h.dom_id = d.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($domainName, $ipAddress) = @{$row};
    $ipAddress ||= $defaultIp;

    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  __populateClientsFromDomains();
}

## Plesk 6.0 - 8.0

sub __init60 {
  __populateIpsFromIPAddressesTable();
  __populateClientIpPool();
  __populateDomains();
}

## common

sub __getDefaultIp {
  unless ($defaultIp) {
    my $sql = "SELECT val FROM misc WHERE param = 'NameVirtualHost'";
    $dbh->execute($sql);
    my @row = @{$dbh->fetchrow()};
    $defaultIp = $row[0];
    $dbh->finish();
  }
  return $defaultIp;
}

sub __populateIpsFromList {
  my ($sql) = @_;

  $ipType{$defaultIp} = "shared";

  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my @row = @{$row};
    my ($ip) = $row[0];
    $ipType{$ip} = $ip eq $defaultIp ? "shared" : "exclusive";
  }
  $dbh->finish();
}

sub __populateIpsFromDns {
  __populateIpsFromList("SELECT DISTINCT(val) FROM dns_recs WHERE type = 'A'");
}

sub __populateClientsFromDomains {
  my %clientToIp;

  my $sql = "SELECT c.login, d.name FROM clients c, domains d WHERE d.cl_id = c.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($clientName, $domainName) = @{$row};

    if (!exists($clientDomains{$clientName})) {
      $clientDomains{$clientName} = [];
    }
    push @{$clientDomains{$clientName}}, $domainName;

    if (!exists($clientToIp{$clientName})) {
      $clientToIp{$clientName} = {};
    }
    $clientToIp{$clientName}->{$domainIp{$domainName}} = 1;
  }
  $dbh->finish();

  for my $clientName (keys %clientToIp) {
    @{$clientIpPool{$clientName}} = keys %{$clientToIp{$clientName}};
  }
}

sub __populateIpsFromIPAddressesTable {
  my $sql = "SELECT ip_address, type FROM IP_Addresses";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    $ipType{$row->[0]} = $row->[1];
  }
  $dbh->finish();
}

sub __populateClientIpPool {
  my $sql = "SELECT c.login, p.ip_address FROM clients c "
    . "LEFT JOIN Repository r ON c.pool_id = r.rep_id "
    . "LEFT JOIN IP_Addresses p ON r.component_id = p.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($clientName, $ipAddress) = @{$row};
    if (!exists $clientIpPool{$clientName}) {
      $clientIpPool{$clientName} = [];
    }
    if ($ipAddress) {
      push @{$clientIpPool{$clientName}}, $ipAddress;
    }
  }
  $dbh->finish();
}

sub __populateDomains {
  my $domainNameColumn = 'name';
  if (PleskVersion::atLeast(7, 1, 0)) {
    $domainNameColumn = 'displayName';
  }

  my $sql = "SELECT d.$domainNameColumn, p.ip_address FROM domains d, IP_Addresses p, hosting h "
    . "WHERE p.id = h.ip_address_id AND h.dom_id = d.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($domainName, $ipAddress) = @{$row};
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  $sql = "SELECT d.$domainNameColumn, p.ip_address FROM domains d, IP_Addresses p, forwarding f "
    . "WHERE p.id = f.ip_address_id AND f.dom_id = d.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($domainName, $ipAddress) = @{$row};
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  $sql = "SELECT $domainNameColumn FROM domains WHERE htype = 'none'";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    $domainIp{$row->[0]} = undef;
  }
  $dbh->finish();

  $sql = "SELECT d.$domainNameColumn, c.login FROM clients c LEFT JOIN domains d ON d.cl_id = c.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($domainName, $clientName) = @{$row};
    if (!exists($clientDomains{$clientName})) {
      $clientDomains{$clientName} = [];
    }
    if ($domainName) {
      push @{$clientDomains{$clientName}}, $domainName;
    }
  }
  $dbh->finish();
}

1;

# Local Variables:
# mode: cperl
# cperl-indent-level: 2
# indent-tabs-mode: nil
# tab-width: 4
# End:
