#!/usr/bin/perl

use strict;

# Prefer internal modules over external
use lib '.';

use PleskX;
use CompatArgParse;
use Storage::Storage;
use DumpStatus;
use Logging;
use XmlNode;

use sigtrap qw(die normal-signals);
use POSIX qw(:signal_h);

sub printHelp
{
  print <<HELP;

Usage:
  $0 <command> [<options>]

Commands:
  -s |--get-status           get status of the host: is this agent applicable
  -dc|--dump-accounts=<list> comma-separated list of clients to dump
  -dd|--dump-domains=<list>  comma-separated list of domains to dump
  -da|--dump-all             dump all objects on hosts
  -lc|--get-content-list     get list of content files
  -h |--help                 this help

Options:
      --server               dump server settings (may be used as independent command)
  -nc|--no-content           make a shallow structure dump
  -co|--configuration-only   dump only configuration
  -om|--only-mail            dump only mail content
  -nz|--no-compress          do not compress content files

  -o |--output=<file>        'tar' the dump files to the single file

  -ss|--split-size=<size>    Split the dump to the parts of <size>.
                             Additional parts have .001, .002... suffixes.
                             Dump is not split if such option is not given.

  -d |--description=<text>   Add description to the dump

  -v |--verbosity=<1-5>      Verbosity level. Default 5

HELP
}

sub main {

  my $sigset = POSIX::SigSet->new;
  sigprocmask(SIG_SETMASK, $sigset);

  my %arg_opts = ('--help|-h'=>'',
                  '--get-status|-s'=>'',
                  '--dump-accounts|-dc'=>'s',
                  '--dump-domains|-dd'=>'s',
                  '--dump-all|-da'=>'',
                  '--get-content-list|-lc'=>'',
                  '--no-content|-nc'=>'',
                  '--only-mail|-om'=>'',
                  '--no-compress|-nz'=>'',
                  '--configuration-only|-co'=>'',
                  '--server'=>'',
                  '--output|-o'=>'s',
                  '--split-size|-ss'=>'s',
                  '--description|-d'=>'s',	
                  '--verbosity|-v'=>'s'
                 );

  my $args = CompatArgParse::parse(\@ARGV,\%arg_opts);

  if (exists $args->{'help'}) {
    printHelp();
    return 0;
  }

  if (exists $args->{'verbosity'}) {
    Logging::setVerbosity($args->{'verbosity'});
  }

  if (exists $args->{'get-status'}) {
    my $root = XmlNode->new('agent-status');

    my @problems = PleskX::checkHost();

    if (@problems) {
      $root->addChild(XmlNode->new('wrong-platform', 'content' => (join "\n", @problems)));
    }

    $root->serialize(\*STDOUT);

    return 0;
  }

  my $do_gzip = exists $args->{'no-compress'} ? undef : 1;

  my $storage;
  my $status;
  if (exists $args->{'output'}) {
    # Backup

    $storage = Storage::Storage::createMimeStorage($do_gzip, $args->{'output'}, $args->{'split-size'});

    #FIXME: finish backup-dump-status
    $status = DumpStatus::createBackup();
  } else {
    # Migration
    my $workDir = AgentConfig::cwd();

    $storage = Storage::Storage::createFileStorage($do_gzip, $workDir, $args->{'split-size'});

    $status = DumpStatus::createMigration($workDir . "/dumping-status.xml");
  }

  if (exists $args->{'get-content-list'}) {
    print $storage->getContentList();
    return 0;
  }

  # in PleskX.pl shared agents code is always in the current directory

  my $agent = PleskX->new($storage, $status, AgentConfig::cwd());

  if (exists $args->{'no-content'}) {
    $agent->setDumpType($PleskX::SHALLOW);
  }
  if (exists $args->{'configuration-only'}) {
    $agent->setDumpType($PleskX::CONFIGURATION);
  }
  if (exists $args->{'only-mail'}) {
    $agent->setDumpType($PleskX::ONLY_MAIL);
  }
  if (exists $args->{'description'}) {
    $agent->setDescription($args->{'description'});
  }

  if (exists $args->{'dump-accounts'}) {
    my $value = $args->{'dump-accounts'};
    if ($value eq "-") {
      $value = <STDIN>;
      chomp $value;
    }
    my @clients = split(/\s*,\s*/, $value);
    Logging::debug("Selected clients: " . join (", ", @clients));
    $agent->selectClients(@clients);
  }

  if (exists $args->{'dump-domains'}) {
    my $value = $args->{'dump-domains'};
    if ($value eq "-") {
      $value = <STDIN>;
      chomp $value;
    }
    my @domains = split(/\s*,\s*/, $value);
    Logging::debug("Selected domains: " . join (", ", @domains));
    $agent->selectDomains(@domains);
  }

  if (exists $args->{'dump-all'}) {
    Logging::debug("Selected all objects");
    $agent->selectAll();
  }

  if (exists $args->{'server'}) {
    $agent->selectServerSettings();
  }

  my $res = $agent->dump();
  $status->finish();
  return $res;
}

exit(main());

# Local Variables:
# mode: cperl
# cperl-indent-level: 2
# indent-tabs-mode: nil
# tab-width: 4
# End:
