package generic;

use Apache::ConfigFile;
use Mail::Qmail;
use Mail::Sendmail;

use strict;
use warnings;

require "agent.include.pl";

use vars qw|@ISA|;

sub new {
  my $self = {};
  bless($self, shift);
  $self->_init(@_);
  return $self;
}

sub _init {
  my ($self, $storage, $status, $dbpassword) = @_;

# Information dumps.
#
# apache_dump - contain information about virtual hosts gathered from apache configuration files
#
# Format:
#           <domain name> => Apache::ConfigFile object.
#
# apache_ip_dump - contain information about conformity virtual hosts and their ip addresses
#
# Format:
#           <domain name> => <ip address>
#
# qmail_dump - contain information about mailnames, domains, redirects and so on gathered from qmail conf. files
#
# Format:
#
#           <domain name> =>  { <mail name> => { directory => value, password => value  etc.}
#

  $self->{apache_dump} = {};
  $self->{apache_ip_dump} = {};

  $self->{qmail_dump} = {};
  $self->{sendmail_dump} = {};

  $self->{db_dump} = {};

  $self->{storage} = $storage;
  $self->{dump_status} = $status;

  $self->{db_password} = $dbpassword;

  $self->constructApacheDump($self->mainApacheConfig(AgentConfig::apachectlBin())) if AgentConfig::apachectlBin();
  $self->constructApacheDump($self->mainApacheConfig(AgentConfig::apache2ctlBin())) if AgentConfig::apache2ctlBin();
  $self->constructQmailDump(Mail::Qmail->new) if AgentConfig::qmailBin();
  $self->constructSendmailDump(Mail::Sendmail->new) if AgentConfig::sendmailBin();

  if (AgentConfig::psqlBin() || AgentConfig::mysqlBin()) {
    my $default_domain_name = `hostname -f`;
    chomp $default_domain_name;
    $self->{db_dump}->{$default_domain_name} = 1;
  }

# Join mail dumps together

  my ($inc, $k, $v, %temporary);
  foreach my $mail ($self->{qmail_dump}, $self->{sendmail_dump}) {
    while (($k, $v) = each %$mail) {
      next if $k eq 'localhost';
      $self->{mail_dump}->{$k} = $v;
    }
  }

# Hosting types.
#
# full_hosting - domain has physical hosting, mail and databases.
# mail_hosting - domain has only mail.
# db_hosting   - domain has only databases.
# web_hosting  - domain has only physical hosting.
# mail_web_hosting - domain has physical hosting and mail.
# db_web_hosting - domain has physical hosting and databases.
#

  $self->{full_hosting} = {};
  $self->{mail_hosting} = {};
  $self->{db_hosting} = {};
  $self->{web_hosting} = {};
  $self->{mail_web_hosting} = {};
  $self->{db_web_hosting} = {};

#  Merge all hash key

  $self->{all_domains} = [];

  foreach $inc ($self->{apache_dump}, $self->{mail_dump}, $self->{db_dump}) {
    while (($k, $v) = each %$inc) {
       next if exists $temporary{$k};
       $temporary{$k} = 1;
    }
  }

  @{$self->{all_domains}} = keys %temporary;

# Sorting domains by hosting type

  foreach my $dom (@{$self->{all_domains}}) {
    if (exists $self->{apache_dump}->{$dom} and exists $self->{mail_dump}->{$dom} and exists $self->{db_dump}->{$dom}) {
      $self->{full_hosting}->{$dom} = 1;
    }elsif(exists $self->{apache_dump}->{$dom} and exists $self->{mail_dump}->{$dom} and not exists $self->{db_dump}->{$dom}) {
      $self->{mail_web_hosting}->{$dom} = 1;
    }elsif(exists $self->{apache_dump}->{$dom} and not exists $self->{mail_dump}->{$dom} and exists $self->{db_dump}->{$dom}) {
      $self->{db_web_hosting}->{$dom} = 1;
    }elsif(exists $self->{apache_dump}->{$dom} and not exists $self->{mail_dump}->{$dom} and not exists $self->{db_dump}->{$dom}) {
      $self->{web_hosting}->{$dom} = 1;
    }elsif(not exists $self->{apache_dump}->{$dom} and exists $self->{mail_dump}->{$dom} and not exists $self->{db_dump}->{$dom}) {
      $self->{mail_hosting}->{$dom} = 1;
    }elsif(not exists $self->{apache_dump}->{$dom} and not exists $self->{mail_dump}->{$dom} and exists $self->{db_dump}->{$dom}) {
      $self->{db_hosting}->{$dom} = 1;
    }
  }
}

sub constructApacheDump {
  my ($self, $config) = @_;
  my (@files, @new_files, $file, $tmp);
  open(CONF, "< $config");
  while(<CONF>) {
    next if ($_=~/^#/);
    if ($_=~/Include (.*)/) {
      push @files, $1;
    }
  }
  close(CONF);

  foreach my $temp (@files){
    if ($temp=~/(.*)\/\[\^\.\#\]\*$/){
      opendir(NCD, $1);
      my @new_confs = readdir(NCD);
      closedir(NCD);
      chomp @new_confs;
      $temp = undef;
      foreach $tmp (@new_confs){
        push @new_files, $1."/".$tmp;
      }
    }elsif($temp=~/(.*)\*\.conf/){
      opendir(NCD, $1);
      my @new_one = readdir(NCD);
      closedir(NCD);
      chomp @new_one;
      $temp = undef;
      foreach $tmp (@new_one) {
        push @new_files, $1."/".$tmp;
      }
    }else{
      push @new_files, $temp;
    }
  }

  foreach $file (@new_files) {
    next unless (-e $file);
    $self->constructApacheUnitDump($file);
    $self->constructApacheDump($file);
  }
}

sub constructApacheUnitDump {
  my ($self, $config_unit)=@_;
  my $ac = Apache::ConfigFile->read($config_unit);

  my @vhkeys = $ac->cmd_context('VirtualHost');

  foreach my $vhkey (@vhkeys) {
    next if ($vhkey=~/\_+|\*+/);
    foreach my $vh ($ac->cmd_context(VirtualHost => $vhkey)) {
      my $sname = $vh->cmd_config('ServerName');

      $sname=~s/:[0-9]+//g;
      $self->{apache_dump}->{$sname} = $vh;

      $vhkey=~s/:[0-9]+//g;
      $self->{apache_ip_dump}->{$sname} =  $vhkey;
    }
  }
}

sub constructQmailDump {
  my ($self, $qmail) = @_;

  my @info = $qmail->get_info_from_assign();

  foreach my $inf (@info) {
    $self->{qmail_dump}->{$inf->{domain}} = { $inf->{mailname} => {'directory' => $inf->{directory}, 'password' => $qmail->get_passwd_by_dir($inf->{directory})}};
  }

}

sub constructSendmailDump {
  my ($self, $sendmail) = @_;
  my @local_domains = $sendmail->getLocalDomains();

  foreach my $domain (@local_domains) {
    $self->{sendmail_dump}->{$domain} = $sendmail->collectMailInfo($sendmail->getDomainMails($domain));
  }
}

sub selectDomains {
  my ($self, @domains) = @_;
  # may be do some checks
  $self->{domains} = \@domains;
}

sub selectAll {
 my ($self) = @_;
 $self->{domains} = $self->{all_domains} ;
}

sub dump {
  my ($self) = @_;
  my $root = XmlNode->new('migration-dump', 'attributes' => {'agent-name' => 'generic'});

  if(exists $self->{domains}){

    $self->{dump_status}->start(0, scalar(@{$self->{domains}}));

    foreach my $domain (@{$self->{domains}}) {
      $root->addChild($self->makeDomain($domain));
    }

    $self->{dump_status}->finishObjects();

  }else{
    die "No objects to dump found";
  }

  return $self->{storage}->finish($root);
}

sub makeDomainNodeFromWeb {
  my ($self, $domain_name) = @_;
  my $vh = $self->{apache_dump}->{$domain_name};
  my $server_alias = $vh->{'cmd_context'}->{'ServerAlias'}->[0]->[0];
  my $item = XmlNode->new('domain', 'attributes' => {'name' => $domain_name,'www' => (defined($server_alias)) ?  'true' : 'false' });

  $self->makeDomainIpNode($item, $self->{apache_ip_dump}->{$domain_name});

  $item->addChild(XmlNode->new('status', 'children' => [XmlNode->new('enabled')]));

  return $item;
}

sub makeDomainNodeFromMail {
  my ($self, $domain_name) = @_;

  my $item = XmlNode->new('domain', 'attributes' => {'name' => $domain_name});

  $item->addChild(XmlNode->new('status', 'children' => [XmlNode->new('enabled')]));

  return $item;

}

sub makeDomainNodeFromDb {
  my ($self, $domain_name) = @_;

  my $item = XmlNode->new('domain', 'attributes' => {'name' => $domain_name});
  my $defaultipaddr = `ifconfig eth0 | perl -ne "if(/inet addr:([\\d.]+)/) { print \\\$1; }"`;
  chomp $defaultipaddr;

  $self->makeDomainIpNode($item, $defaultipaddr);

  $item->addChild(XmlNode->new('status', 'children' => [XmlNode->new('enabled')]));

  return $item;
}

sub makeDomain {
  my ($self, $domain_name) = @_;

  my ($domain);

  if(exists $self->{db_hosting}->{$domain_name}) {
    $domain = $self->makeDomainNodeFromDb($domain_name);
    $self->makeDatabaseNode($domain, $domain_name);
  }elsif(exists $self->{mail_hosting}->{$domain_name}){
    $domain = $self->makeDomainNodeFromMail($domain_name);
    $self->makeMailSystemNode($domain, $domain_name);
  }else{
    $domain = $self->makeDomainNodeFromWeb($domain_name);
    $self->makeMailSystemNode($domain, $domain_name) unless (exists $self->{db_web_hosting}->{$domain_name} or exists $self->{web_hosting}->{$domain_name});
    $self->makePhostingNode($domain, $domain_name);
  }

  return $domain;

}

sub makeDomainIpNode {
  my ($self, $parent, $ip) = @_;
  $parent->addChild(XmlNode->new('ip', 'children' => [XmlNode->new('ip-type', 'content' => 'shared'), XmlNode->new('ip-address', 'content' => $ip)]));
}

sub makePhostingNode {
  my ($self, $parent, $domain_name) = @_;

  my $vh = $self->{apache_dump}->{$domain_name};
  my $item = XmlNode->new('phosting');
  my ($cid, $cid_cgi, $cid_ssl);

  #contents
  my $doc_root = $vh->cmd_config('DocumentRoot');

  if(-d $doc_root) {
    $cid = $self->addTar("$domain_name.httpdocs", "directory" => $doc_root);
    $item->setAttribute('cid_docroot', $cid);
  }

  my @script_alias = $vh->cmd_config_array('ScriptAlias');

  for my $script (@script_alias) {
    if ($script->[0]=~/cgi-bin/) {
      $cid_cgi = $self->addTar("$domain_name.cgi", "directory" => $script->[1]);
      $item->setAttribute('cid_cgi', $cid_cgi);
    }
  }

#  if (ref($objSsl) eq "ConfigFile"){
#    my $docRootSsl = $objSsl->cmd_config('DocumentRoot');
#    $cidSsl = makeDumpFile("$dumpDir/$serverName\_ssl",$docRootSsl);
#    $item->{'ATTRIBUTE'}->('https','true');
#    $item->{'ATTRIBUTE'}->('cid_docroot_ssl',$cidSsl);
#  }

  # sysuser

  $self->_makeSysUserNode($item, $doc_root);

  # scripting

  $self->makeScriptingNode($item, $vh, $doc_root);

  $parent->addChild($item);
}

sub makeScriptingNode {
  my ($self, $parent, $vh, $doc_root) = @_;
  my $item = XmlNode->new('scripting');

  my $scripts = $vh->cmd_context(Directory => $doc_root);

  if (ref($scripts)=~/ConfigFile/){
    if(ref($scripts->cmd_context(IfModule => 'mod_python.c'))=~/ConfigFile/){
      $item->setAttribute('python','true');
    }

    #cgi
    my @cgi_alias = $vh->cmd_config_array('ScriptAlias');
    for my $script (@cgi_alias){
      if ($script->[0]=~/cgi-bin/){
        $item->setAttribute('cgi','true');
      }
    }

    #php
    my $php = $scripts->cmd_context(IfModule => 'mod_php5.c'); #need to think about other versions of mod_php.
    if (ref($php)=~/ConfigFile/) {
      my @php_opts = $php->cmd_config_array('php_admin_flag');
      for my $php_opt (@php_opts){
        if ($php_opt->[1] eq "on"){
          $item->setAttribute('php','true');
        }
      }
    }

    #ssi
    my @opts = $scripts->cmd_config_array('Options');
    for my $opt (@opts){
      if ($opt->[0] eq "+Includes"){
        $item->setAttribute('ssi','true');
      }
    }

    #mod_perl && asp
    my @mod_perl_opts = $scripts->cmd_context(IfModule => 'mod_perl.c');
    for my $mod_perl_opt (@mod_perl_opts){
      if (ref($mod_perl_opt) eq "ConfigFile"){
        if ($mod_perl_opt->cmd_config('Files') eq "\~ \(\\.asp\$\)") { $item->{'ATTRIBUTE'}->('asp','true')};
        if ($mod_perl_opt->cmd_config('Files') eq "\~ \(\\.pl\$\)") { $item->{'ATTRIBUTE'}->('perl','true')};
      }
    }
  }

  $parent->addChild($item);

}

sub _makeSysUserNode {
  my ($self, $parent, $doc_root) = @_;
  my @dir_props = stat($doc_root);
  my $user_params = $self->parsePasswdShadow($dir_props[4]);
  my $item = XmlNode->new('sysuser');
  my $user_name = $user_params->{$dir_props[4]}->{'name'};

  $item->setAttribute('name', $user_name);
  $item->setAttribute('shell', $user_params->{$dir_props[4]}->{'shell'});

  $item->addChild(makePasswordNode($user_params->{$dir_props[4]}->{'password'}, 'encrypted'));
  $parent->addChild($item);
}

sub parsePasswdShadow {
  my ($self, $uid)=@_;

  my $passwdReader = makeFileParser("/etc/passwd");
  my $shadowReader = makeSafeFileParser("/etc/shadow");

  my (%passwd, %shadow);
  if (defined($passwdReader)) {
    %passwd = %{$passwdReader->{'PARSE'}->('KEYSEPARATOR' => ':', 'VALUESEPARATOR' => ':')};
  } else {
    return undef;
  }

  if (defined($shadowReader)) {
    %shadow = %{$shadowReader->{'PARSE'}->('KEYSEPARATOR' => ':', 'VALUESEPARATOR' => ':')};
  }

  my ($user,$paramsPtr,@params,$password,$shell,@shadowParams,%userParams);

  while (($user,$paramsPtr) = each %passwd){
    @params = @{$paramsPtr};
    if($uid == $params[1]){
      $shell = $params[5];
      if (%shadow && defined($shadow{$user})){
        @shadowParams = @{$shadow{$user}};
        $password = $shadowParams[0];
      }
      $userParams{$uid} = {'name' => $user,'shell' => $shell, 'password' => $password};
    }
  }

  return \%userParams;
}

sub makeMailSystemNode {
  my ($self, $parent, $domain_name) = @_;
  my $mailsystem = XmlNode->new('mailsystem');

  my %domain_mail = %{$self->{mail_dump}->{$domain_name}};
  # Add status
  $mailsystem->addChild(XmlNode->new('status', 'children' => [XmlNode->new('enabled')]));
  if(exists $self->{mail_dump}->{$domain_name}) {
    foreach my $key (keys %domain_mail) {
      $self->makeMailUserNode($mailsystem, $key, $domain_mail{$key}, $domain_name);
    }
  }
  $parent->addChild($mailsystem);
}

sub makeMailUserNode {
  my ($self, $parent, $mailname, $opts, $domain) = @_;
  my $mailuser = XmlNode->new('mailuser', 'attributes' => {'name' => $mailname, 'mailgroup-enabled' => (exists $opts->{mailgroup}) ? 'true' : 'false'});
  $mailuser->addChild(makePasswordNode($opts->{password},'encrypted')) if exists $opts->{password};
  if (exists $opts->{directory}) {
    my $mailbox = XmlNode->new('mailbox', 'attributes' => {'type' => 'mdir'});
    my $cid = $self->addTar("$mailname\@$domain.mdir", 'directory' => $opts->{directory});
    if ($cid) {
      $mailbox->setAttribute('cid', $cid);
    }
    $mailuser->addChild($mailbox);
  }
 $parent->addChild($mailuser);
}

sub makeDatabaseNode {
  my ($self, $root) = @_;

  my ($usermysqlpasswd, $usermysql) = $self->getUserDbConnect();

  return unless $self->validConnection($usermysql);

  my $sql = "SHOW DATABASES";
  my @dbs;
  if ($usermysql->{'EXECUTE'}->($sql)) {
	while (my $r = $usermysql->{'FETCHROW'}->()) {
	  my $db = $r->[0];
      Logging::info("Database: ".$db);
	  next if $db eq "mysql" or $db eq "test";
	  push @dbs, $db;
	}
  }
  $usermysql->{'FINISH'}->();

  foreach my $db (@dbs) {
    my $dbnode = XmlNode->new("database", "attributes" => { "name" => $db, "type" => "mysql" });
    my $cid = $self->addDb("$db.mysql", "name" => $db, "type" => "mysql", "user" => "root" , "password" => $usermysqlpasswd);
    $dbnode->setAttribute('cid', $cid) if $cid;
    $dbnode->setAttribute('version', Db::MysqlUtils::getVersion());

	my %dbusers;

	$sql = "SELECT user.Host, user.User, user.Password FROM user, db ".
	  " WHERE db.Select_priv = 'Y' AND db.Host = user.Host AND db.Db = '$db' ".
	  " AND db.User = user.User AND user.User <> ''";
	if ($usermysql->{'EXECUTE'}->($sql)) {
	  while (my $h = $usermysql->{'FETCHHASH'}->()) {
		unless (exists $dbusers{$h->{'User'}}) {
		  $dbusers{$h->{'User'}} = {};
		  $dbusers{$h->{'User'}}->{'accesshost'} = [];
		}
		push @{$dbusers{$h->{'User'}}->{'accesshost'}}, $h->{'Host'};

		unless (exists $dbusers{$h->{'User'}}->{'password'} or $h->{'Password'} eq '') {
		  $dbusers{$h->{'User'}}->{'password'} = $h->{'Password'};
		}
	  }
	}
	$usermysql->{'FINISH'}->();

	foreach my $dbuser (keys(%dbusers)) {
	  my $dbusernode = XmlNode->new("dbuser",
							   "attributes" => { "name" => $dbuser },
							   "children" => [makePasswordNode($dbusers{$dbuser}->{'password'})]);
	  foreach my $host (@{$dbusers{$dbuser}->{'accesshost'}}) {
		$dbusernode->addChild(XmlNode->new("accesshost", "content" => $host));
	  }
	  $dbnode->addChild($dbusernode);
	}
	$root->addChild($dbnode);
  }

  return $root;
}

sub getUserDbConnect {
  my ($self) = @_;

  my $usermysql;

#  Logging::info("Trying to connect to user's DB without password...");
#  $usermysql = getUserDbConnect2("");
#  return ("", $usermysql) if validConnection($usermysql);

  Logging::info("Trying to connect to user's DB with users password...");
  $usermysql = getUserDbConnect2($self->{db_password});
  return ($self->{db_password}, $usermysql) if validConnection($usermysql);
}

sub validConnection {
  my ($self, $connect) = @_;
  return ref($connect) =~ /HASH/ && ref($connect->{'EXECUTE'}) =~ /CODE/;
}

sub getUserDbConnect2 {
  my ($self, $password) = @_;

  return getDbConnect('mysql', "root", $password, "mysql", "localhost", 1);
}

sub addTar {
  my $self = shift;
  return $self->{storage}->addTar(@_);
}

sub addDb {
  my $self = shift;
  return $self->{storage}->addDb(@_);
}

sub mainApacheConfig {
  my ($self, $apachectl) = @_;

  my $config;

  open (CTL, "$apachectl -V |");
  while (<CTL>) {
    chomp;
    if ($_=~/\w+\s+SERVER_CONFIG_FILE=(.+)/) {
      $config = $1;
      $config=~s/"//g;
    }
  }
  close CTL;

  return $config;
}

1;
