package Storage::FileStorage;

use strict;
use Storage::Storage;
use Logging;

use POSIX;
use IPC::Run;

use Storage::Splitter;
use Storage::Counter;

use vars qw|@ISA|;

@ISA = qw|Storage::Storage|;

sub _init {
  my ($self, %options) = @_;

  $self->SUPER::_init(%options);
  $self->{split_size} = $options{split_size};
  $self->{gzip_bundle} = $options{gzip_bundle};
  $self->{output_dir} = $options{output_dir};

  $self->{last_used_id} = 0;

  $self->{bundle_files} = [];

  $self->{unpacked_size} = 0;
  $self->{packed_size} = 0;

  Logging::info("-" x 60);
  Logging::info("FILE storage initialized.");
  Logging::info("Base directory: $self->{output_dir}");
  Logging::info("Gzip bundles: " . ($self->{gzip_bundle} ? "yes" : "no"));
  Logging::info("Bundle split size: " . ($self->{split_size} || "do not split"));
  Logging::info("-" x 60);
}

#
# Checks the validity of proposed id: it should not be too long.
#

sub getFileNameIdFromId {
  my ($self, $id) = @_;

  my $maxLength = &POSIX::PATH_MAX;
  if ($self->{split_size}) {
    $maxLength -= 4;
  }

  my $destFile = $self->{output_dir} . "/" . $id;
  if (length($destFile) > $maxLength) {
    $id = $self->{last_used_id}++;
    $destFile = $self->{output_dir} . "/" . $id;
  }

  if ($self->{gzip_bundle}) {
    $id .= ":gzipped";
  }

  return ($destFile, $id);
}

sub getBundleExecutor {
  my ($bundle) = @_;

  return sub {
    my $exec = $bundle->run();
    binmode STDOUT;

    my $block;
    my $blocklen;
    while ($blocklen = sysread($exec, $block, 65536)) {
       my $offset = 0;
       do {
         my $written = syswrite(STDOUT, $block, $blocklen, $offset);
         $offset += $written;
         $blocklen -= $written;
       } while ($blocklen != 0);
    }

    $bundle->cleanup();

    POSIX::_exit(0);
  };
}

sub executeAndSave {
  my ($self, $destFile, $bundle) = @_;

  my $unpackedSize;

  #allocating filehandle for creating pipe from subprocess
  my $newhandle = POSIX::open("/dev/null", O_RDWR, 0666);

  my @cmd;
  push @cmd, getBundleExecutor($bundle);
  push @cmd, "|", \&Storage::Counter::run, "$newhandle>", \$unpackedSize,
    init => sub {Storage::Counter::init($newhandle)};

  if ($self->{gzip_bundle}) {
    #FIXME check gzip
    push @cmd, "|", ["gzip"];
  }

  my $files;

  my $newhandle2 = POSIX::open("/dev/null", O_RDWR, 0666);

  push @cmd, "|", \&Storage::Splitter::run, "$newhandle2>", \$files,
    init => sub {Storage::Splitter::init_process($newhandle2, $self->{split_size}, $destFile)};

  my $h = IPC::Run::harness(@cmd);

  if (!$h->run()) {
    Logging::error("Failed to pack files\n");
    POSIX::close($newhandle);
    POSIX::close($newhandle2);
    return;
  }

  POSIX::close($newhandle);
  POSIX::close($newhandle2);

  if ($unpackedSize =~ /ERR\s(.*)/) {
    Logging::error("Unable to pipe data through filter: $1");
    return;
  }

  $self->{unpacked_size} += $unpackedSize;

  foreach my $line (split/\n/, $files) {
    my ($file_name, $file_size) = split / /, $line;
    push @{$self->{bundle_files}}, [$file_name, $file_size];
    $self->{packed_size} += $file_size;
  }

  return 1;
}

sub addDb {
  my ($self, $proposedId, %options) = @_;

  my ($destFile, $id) = $self->getFileNameIdFromId($proposedId, $self->{gzip_bundle});
  Logging::debug("DB bundle. id=$id, destFile=$destFile");

  my $bundle = Storage::Bundle::createDbBundle(%options, 'gzip' => $self->{gzip_bundle});
  return unless $bundle;
  return $id if $self->executeAndSave($destFile, $bundle);
}

sub addTar {
  my ($self, $proposedId, %options) = @_;

  return unless -d $options{'directory'};

  my ($destFile, $id) = $self->getFileNameIdFromId($proposedId);
  Logging::debug("Tar bundle. id=$id, destFile=$destFile");

  my $bundle = Storage::Bundle::createTarBundle(%options, 'gzip' => $self->{gzip_bundle});
  return unless $bundle;
  return $id if $self->executeAndSave($destFile, $bundle);
}

sub _getContentListFileName {
  my ($self) = @_;
  return $self->{output_dir} . "/content-list.xml";
}

sub finish {
  my ($self, $descriptor) = @_;

  my $dumpFile = $self->{output_dir} . "/dump.xml";

  Logging::info("Writing dump file: $dumpFile");

  open DUMPFILE, "> $dumpFile";
  $descriptor->{'PRINT_TREE'}->(\*DUMPFILE);
  close DUMPFILE;

  open CONTENT_FILE, ">" . $self->_getContentListFileName();
  print CONTENT_FILE "<contentlist unpacked='$self->{unpacked_size}' size='$self->{packed_size}'>\n";
  foreach my $ptrRow (@{$self->{bundle_files}}) {
    my $file_name = $ptrRow->[0];
    my $file_size = $ptrRow->[1];
    print CONTENT_FILE "  <file size='$file_size'>$file_name</file>\n";
  }
  print CONTENT_FILE "</contentlist>\n";
  close CONTENT_FILE;
  return 0;
}

sub getContentList {
  my ($self) = @_;

  open CONTENT_FILE, $self->_getContentListFileName();
  my $s =  join "", <CONTENT_FILE>;
  close CONTENT_FILE;
  return $s;
}

1;

# Local Variables:
# mode: cperl
# cperl-indent-level: 2
# indent-tabs-mode: nil
# tab-width: 4
# End:
